#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/oshtml.cpp,v 1.2 1999/05/17 02:52:22 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  oshtml.c - interface layer from TADS to TADS-HTML front-end
Function
  This layer takes the place of the OS layer normally used in the TADS
  runtime.  Rather than providing an OS interface, this layer provides
  an interface the HTML front-end.
Notes
  
Modified
  11/02/97 MJRoberts  - Creation
*/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

/* include TADS interfaces that we'll need */
#include <os.h>
#include <run.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTMLSYS_H
#include "htmlsys.h"
#endif
#ifndef HTMLPRS_H
#include "htmlprs.h"
#endif

/* status mode flag */
static int status_mode_ = 0;

/* HTML mode flag */
static int html_mode_ = 0;

/*
 *   Buffer containing score information for status line 
 */
static textchar_t status_score_buf_[256];

/*
 *   System frame object.  The system-dependent code must initialize this
 *   by calling oshtml_set_frame() after the frame object has been
 *   created.  
 */
static CHtmlSysFrame *app_frame_ = 0;

/*
 *   Set the system frame object.  The system-specific code must call this
 *   routine to initialize the frame object as soon as the frame object
 *   has been created.  (The system-specific code is responsible for
 *   creating the frame object in the first place, too.)  
 */
void oshtml_set_frame(CHtmlSysFrame *frame)
{
    app_frame_ = frame;
}

/*
 *   Get the app frame object 
 */
CHtmlSysFrame *oshtml_get_frame()
{
    return app_frame_;
}

/*
 *   Enter HTML mode 
 */
void os_start_html()
{
    CHtmlParser *parser;
    
    /* if there's no app frame, there's nothing more to do */
    if (app_frame_ == 0)
        return;

    /* 
     *   Flush buffered text to the parser.  We don't need to format it
     *   right now, but we do need to parse it, since we're going to
     *   change the parsing mode. 
     */
    app_frame_->flush_txtbuf(FALSE);

    /* note that we're in HTML mode */
    html_mode_ = 1;

    /* turn on HTML parsing */
    parser = app_frame_->get_parser();
    parser->obey_whitespace(FALSE, TRUE);
    parser->obey_markups(TRUE);
}

/*
 *   exit HTML mode 
 */
void os_end_html()
{
    CHtmlParser *parser;

    /* if there's no app frame, there's nothing to do */
    if (app_frame_ == 0)
        return;

    /* flush buffered text to the parser */
    app_frame_->flush_txtbuf(FALSE);

    /* note that we're not in HTML mode */
    html_mode_ = 0;

    /* turn off HTML parsing */
    parser = app_frame_->get_parser();
    parser->obey_whitespace(TRUE, TRUE);
    parser->obey_markups(FALSE);
    parser->obey_end_markups(FALSE);
}

/*
 *   check for user break (control-C, etc) 
 */
int os_break()
{
    /* if there's no app frame, there can't be a break */
    if (app_frame_ == 0)
        return FALSE;

    /* ask the window to check for a break */
    return app_frame_->check_break_key();
}

/* 
 *   display a string in the score area in the status line 
 */
void os_strsc(const char *p)
{
    size_t len;

    /* 
     *   remember the score string - it will be used the next time we
     *   rebuild the status line 
     */
    len = get_strlen(p) * sizeof(textchar_t);
    if (len > sizeof(status_score_buf_))
        len = sizeof(status_score_buf_) - sizeof(status_score_buf_[0]);
    memcpy(status_score_buf_, p, len);
    status_score_buf_[len / sizeof(status_score_buf_[0])] = '\0';
}

/*
 *   internal routine to display HTML tags, regardless of our current
 *   markup obedience mode 
 */
static void oshtml_display_html_tags(const textchar_t *txt)
{
    CHtmlParser *parser;
    int old_obey, old_obey_end;

    /* if there's no app frame, there's nothing to do */
    if (app_frame_ == 0)
        return;

    /* get the parser */
    parser = app_frame_->get_parser();

    /* 
     *   Flush any buffered text to the parser.  We don't need to format
     *   it yet, but we do need to make sure it's fully parsed, since
     *   we're may be changing the parsing mode.  
     */
    app_frame_->flush_txtbuf(FALSE);

    /* obey markups long enough to parse our HTML markup text */
    old_obey = parser->get_obey_markups();
    old_obey_end = parser->get_obey_end_markups();
    parser->obey_markups(TRUE);

    /* display the output */
    app_frame_->display_output(txt);

    /* flush the result to the parser, since we may be changing modes */
    app_frame_->flush_txtbuf(FALSE);

    /* restore markup obedience mode */
    parser->obey_markups(old_obey);
    parser->obey_end_markups(old_obey_end);
}

/*
 *   internal routine to set the status mode 
 */
static void oshtml_status_mode(int flag)
{
    static const textchar_t banner_start[] =
        "<banner id=statusline height=previous border>"
        "<body bgcolor=statusbg text=statustext><b>";
    static const textchar_t banner_score[] = "</b><tab align=right><i>";
    static const textchar_t banner_end[] = "</i><br height=0></banner>";

    /* start or end the status line banner */
    if (flag)
    {
        /* start the banner */
        oshtml_display_html_tags(banner_start);
    }
    else
    {
        /* add the score to the banner */
        oshtml_display_html_tags(banner_score);
        oshtml_display_html_tags(status_score_buf_);
        oshtml_display_html_tags(banner_end);
    }
}

/* 
 *   set the status line mode 
 */
void os_status(int stat)
{
    /* 
     *   If we're in HTML mode, don't do any automatic status line
     *   generation -- let the game do the status line the way it wants,
     *   without any predefined handling from the run-time system 
     */
    if (html_mode_)
        return;

    /* see what mode we're setting */
    switch(stat)
    {
    case 0:
    default:
        /* turn off status line mode */
        status_mode_ = 0;
        oshtml_status_mode(0);
        break;

    case 1:
        /* turn on status line mode */
        status_mode_ = 1;
        oshtml_status_mode(1);
        break;
    }
}

/*
 *   get the status line mode 
 */
int os_get_status()
{
    return status_mode_;
}

/* 
 *   set the score value 
 */
void os_score(int cur, int turncount)
{
    char buf[128];
    
    /* build the default status line score format */
    sprintf(buf, "%d/%d", cur, turncount);

    /* set the score string */
    os_strsc(buf);
}


/* 
 *   read a line of input from the keyboard 
 */
unsigned char *os_gets(unsigned char *buf, size_t bufl)
{
    /* if app frame, we can't get input */
    if (app_frame_ == 0)
        return 0;

    /* if we're in non-HTML mode, establish the input font */
    if (!html_mode_)
        oshtml_display_html_tags("<font face='TADS-Input'>");

    /* 
     *   get input from the app frame - return null if it returns end of
     *   file (i.e., application terminating)
     */
    if (!app_frame_->get_input((textchar_t *)buf, bufl))
        return 0;

    /* if we're in non-HTML mode, close the input font tag */
    if (!html_mode_)
        oshtml_display_html_tags("</font>");
    
    /* success */
    return buf;
}


/* 
 *   read a character from the keyboard 
 */
int os_getc()
{
    /* if the app frame is missing, we can't get input - indicate EOF */
    if (app_frame_ == 0)
    {
        static int eof_cmd;

        /* 
         *   return (0 + CMD_EOF) sequences as we're called; alternate
         *   between the '\000' character and the CMD_EOF character 
         */
        eof_cmd = !eof_cmd;
        return (eof_cmd ? 0 : CMD_EOF);
    }

    /* ask the app frame to get a key */
    return textchar_to_int(app_frame_->wait_for_keystroke(FALSE));
}

/*
 *   Read a character from the keyboard in raw mode.  Because the OS code
 *   handles command line editing itself, we presume that os_getc() is
 *   already returning the low-level raw keystroke information, so we'll
 *   just return that same information. 
 */
int os_getc_raw()
{
    return os_getc();
}


/* wait for a character to become available from the keyboard */
void os_waitc()
{
    /* tell the app frame to wait for a keystroke */
    if (app_frame_ != 0)
        app_frame_->wait_for_keystroke(TRUE);
}

/*
 *   get an event 
 */
int os_get_event(unsigned long timeout_in_milliseconds, int use_timeout,
                 os_event_info_t *info)
{
    /* if there's no app frame, we can't get an event */
    if (app_frame_ == 0)
        return OS_EVT_EOF;

    /* ask the app frame for an event and return the result */
    return app_frame_->get_input_event(timeout_in_milliseconds, use_timeout,
                                       info);
}

/* 
 *   clear the screen 
 */
void oscls()
{
    /* if app frame, we can't do this */
    if (app_frame_ == 0)
        return;

    /* tell the app frame to start a new page */
    app_frame_->start_new_page();
}

/* 
 *   flush any buffered display output 
 */
void os_flush()
{
    /* tell the app frame to flush its buffers */
    if (app_frame_ != 0)
        app_frame_->flush_txtbuf(TRUE);
}

/*
 *   Get highlighting character code.  This returns a character code to
 *   write to the display in order to activate a highlighting mode: mode 1
 *   = turn off boldface, mode 2 = turn on boldface.  A return value of
 *   zero indicates that highlighting is not supported; otherwise, the
 *   return value is the character to write to the display to activate the
 *   given mode.
 *   
 *   This is used for non-HTML mode only.  We'll use the standard control
 *   codes of 1 = boldface off and 2 = boldface on, just like most
 *   character-mode run-times.  
 */
int os_hilite(int mode)
{
    /* control code == mode, by some weird coincidence */
    return mode;
}

/* 
 *   use plain ascii mode for the display 
 */
void os_plain()
{
    /* ignore this -- we can only use the HTML mode */
}

/* 
 *   hide/show graphical mouse, if appropriate 
 */
void os_mouhide()
{
    /* nothing to do - we let Windows manage mouse visibility */
}

void os_moushow()
{
    /* nothing to do - we let Windows manage mouse visibility */
}

/*
 *   printf() equivalent 
 */
void os_printf(const char *fmt, ...)
{
    va_list argptr;

    va_start(argptr, fmt);
    os_vprintf(fmt, argptr);
    va_end(argptr);
}

void os_vprintf(const char *fmt, va_list argptr)
{
    char buf[1024];
    char *p;

    /* apply the formatting */
    vsprintf(buf, fmt, argptr);

    /* check our status mode */
    switch(status_mode_)
    {
    case 0:
        /* display normally */
        break;

    case 1:
        /* display only up to the newline, and eat the rest */
        for (p = buf ; *p && *p != '\n' ; ++p) ;
        if (*p == '\n')
        {
            /* don't display the newline or anything after it */
            *p = '\0';

            /* switch to status mode 2 */
            status_mode_ = 2;
        }
        break;

    case 2:
    default:
        /* 
         *   eat everything in status mode 2 - this is the part after the
         *   initial line of status text 
         */
        return;
    }

    /* if app frame, we can't display anything */
    if (app_frame_ == 0)
        return;

    /* display it */
    app_frame_->display_output(buf);
}

/*
 *   show the MORE prompt and wait for the user to respond
 */
void os_more_prompt()
{
    /* if app frame, we can't do this */
    if (app_frame_ == 0)
        return;

    /* flush the output, and make sure it's displayed */
    app_frame_->flush_txtbuf(TRUE);

    /* display the MORE prompt */
    app_frame_->pause_for_more();
}

/* 
 *   pause prior to exit 
 */
void os_expause()
{
    /* if app frame, we can't do this */
    if (app_frame_ == 0)
        return;

    /* flush the output, and make sure it's displayed */
    app_frame_->flush_txtbuf(TRUE);

    /* display an exit message in the status line, and wait for a key */
    app_frame_->pause_for_exit();
}

/*
 *   printf to debug log window
 */
void oshtml_dbg_printf(const char *fmt, ...)
{
    va_list argptr;

    va_start(argptr, fmt);
    oshtml_dbg_vprintf(fmt, argptr);
    va_end(argptr);
}

void oshtml_dbg_vprintf(const char *fmt, va_list argptr)
{
    char buf[1024];
    char *p;

    /* format the message */
    vsprintf(buf, fmt, argptr);

    /* 
     *   Remove any bold on/off sequences from the buffer.  Bold sequences
     *   are interpreted by the HTML parser as tag open/close sequences,
     *   so they can cause weird problems. 
     */
    for (p = buf ; *p != '\0' ; ++p)
    {
        /* if it's a bold on/off sequence, convert it to a space */
        if (*p == 1 || *p == 2)
            *p = ' ';
    }

    /* display it */
    if (app_frame_ != 0)
    {
        app_frame_->dbg_print(buf);
    }
#ifdef TADSHTML_DEBUG
    else
    {
        /* no app frame - use the system debug console instead */
        os_dbg_sys_msg(buf);
    }
#endif /* TADSHTML_DEBUG */
}

/*
 *   We can ignore title settings from the high-level output formatter
 *   layer, since we parse the title tag ourselves in the underlying HTML
 *   layer.  
 */
void os_set_title(const char *)
{
    /* nothing to do */
}


/*
 *   HTML 4 character set translation.  This translates to the plain ASCII
 *   7-bit character set.  For the HTML interpreter, we only need this for
 *   translations to log files, since we handle the display in the HTML
 *   layer.  
 */
void os_xlat_html4(unsigned int html4_char, char *result, size_t result_len)
{
    /* default character to use for unknown charaters */
#define INV_CHAR "?"

    /* 
     *   Translation table - provides mappings for characters the ISO
     *   Latin-1 subset of the HTML 4 character map (values 128-255).
     *   
     *   Characters marked "(approx)" are approximations where the actual
     *   desired character is not available in the DOS console character
     *   set, but a reasonable approximation is used.  Characters marked
     *   "(approx unaccented)" are accented characters that are not
     *   available; these use the unaccented equivalent as an
     *   approximation, since this will presumably convey more meaning
     *   than a blank.
     *   
     *   Characters marked "(n/a)" have no equivalent (even approximating)
     *   in the DOS character set, and are mapped to spaces.
     *   
     *   Characters marked "(not used)" are not used by HTML '&' markups.  
     */
    static const char *xlat_tbl[] =
    {
        INV_CHAR,                                         /* 128 (not used) */
        INV_CHAR,                                         /* 129 (not used) */
        "'",                                        /* 130 - sbquo (approx) */
        INV_CHAR,                                         /* 131 (not used) */
        "\"",                                       /* 132 - bdquo (approx) */
        INV_CHAR,                                         /* 133 (not used) */
        INV_CHAR,                                     /* 134 - dagger (n/a) */
        INV_CHAR,                                     /* 135 - Dagger (n/a) */
        INV_CHAR,                                         /* 136 (not used) */
        INV_CHAR,                                     /* 137 - permil (n/a) */
        INV_CHAR,                                         /* 138 (not used) */
        "<",                                       /* 139 - lsaquo (approx) */
        INV_CHAR,                                      /* 140 - OElig (n/a) */
        INV_CHAR,                                         /* 141 (not used) */
        INV_CHAR,                                         /* 142 (not used) */
        INV_CHAR,                                         /* 143 (not used) */
        INV_CHAR,                                         /* 144 (not used) */
        "'",                                        /* 145 - lsquo (approx) */
        "'",                                        /* 146 - rsquo (approx) */
        "\"",                                       /* 147 - ldquo (approx) */
        "\"",                                       /* 148 - rdquo (approx) */
        INV_CHAR,                                         /* 149 (not used) */
        "-",                                                /* 150 - endash */
        "--",                                               /* 151 - emdash */
        INV_CHAR,                                         /* 152 (not used) */
        "(tm)",                                     /* 153 - trade (approx) */
        INV_CHAR,                                         /* 154 (not used) */
        ">",                                       /* 155 - rsaquo (approx) */
        INV_CHAR,                                      /* 156 - oelig (n/a) */
        INV_CHAR,                                         /* 157 (not used) */
        INV_CHAR,                                         /* 158 (not used) */
        "Y",                              /* 159 - Yuml (approx unaccented) */
        INV_CHAR,                                         /* 160 (not used) */
        "!",                                                 /* 161 - iexcl */
        "c",                                                  /* 162 - cent */
        "#",                                                 /* 163 - pound */
        INV_CHAR,                                     /* 164 - curren (n/a) */
        INV_CHAR,                                              /* 165 - yen */
        "|",                                                /* 166 - brvbar */
        INV_CHAR,                                       /* 167 - sect (n/a) */
        INV_CHAR,                                        /* 168 - uml (n/a) */
        "(c)",                                       /* 169 - copy (approx) */
        INV_CHAR,                                             /* 170 - ordf */
        "<<",                                                /* 171 - laquo */
        INV_CHAR,                                              /* 172 - not */
        " ",                                             /* 173 - shy (n/a) */
        "(R)",                                        /* 174 - reg (approx) */
        INV_CHAR,                                       /* 175 - macr (n/a) */
        INV_CHAR,                                              /* 176 - deg */
        "+/-",                                              /* 177 - plusmn */
        INV_CHAR,                                             /* 178 - sup2 */
        INV_CHAR,                                       /* 179 - sup3 (n/a) */
        "'",                                        /* 180 - acute (approx) */
        INV_CHAR,                                            /* 181 - micro */
        INV_CHAR,                                       /* 182 - para (n/a) */
        INV_CHAR,                                           /* 183 - middot */
        ",",                                        /* 184 - cedil (approx) */
        INV_CHAR,                                       /* 185 - sup1 (n/a) */
        INV_CHAR,                                             /* 186 - ordm */
        ">>",                                                /* 187 - raquo */
        "1/4",                                             /* 188 - frac14 */
        "1/2",                                             /* 189 - frac12 */
        "3/4",                                     /* 190 - frac34 (approx) */
        "?",                                                /* 191 - iquest */
        "A",                            /* 192 - Agrave (approx unaccented) */
        "A",                            /* 193 - Aacute (approx unaccented) */
        "A",                             /* 194 - Acirc (approx unaccented) */
        "A",                            /* 195 - Atilde (approx unaccented) */
        "A",                                                  /* 196 - Auml */
        "A",                                                 /* 197 - Aring */
        "AE",                                                /* 198 - AElig */
        "C",                                                /* 199 - Ccedil */
        "E",                            /* 200 - Egrave (approx unaccented) */
        "E",                                                /* 201 - Eacute */
        "E",                             /* 202 - Ecirc (approx unaccented) */
        "E",                              /* 203 - Euml (approx unaccented) */
        "I",                            /* 204 - Igrave (approx unaccented) */
        "I",                            /* 205 - Iacute (approx unaccented) */
        "I",                             /* 206 - Icirc (approx unaccented) */
        "I",                              /* 207 - Iuml (approx unaccented) */
        INV_CHAR,                                        /* 208 - ETH (n/a) */
        "N",                                                /* 209 - Ntilde */
        "O",                            /* 210 - Ograve (approx unaccented) */
        "O",                            /* 211 - Oacute (approx unaccented) */
        "O",                             /* 212 - Ocirc (approx unaccented) */
        "O",                            /* 213 - Otilde (approx unaccented) */
        "O",                                                  /* 214 - Ouml */
        "x",                                        /* 215 - times (approx) */
        "O",                            /* 216 - Oslash (approx unaccented) */
        "U",                            /* 217 - Ugrave (approx unaccented) */
        "U",                            /* 218 - Uacute (approx unaccented) */
        "U",                             /* 219 - Ucirc (approx unaccented) */
        "U",                                                  /* 220 - Uuml */
        "Y",                            /* 221 - Yacute (approx unaccented) */
        INV_CHAR,                                      /* 222 - THORN (n/a) */
        INV_CHAR,                                   /* 223 - szlig (approx) */
        "a",                                                /* 224 - agrave */
        "a",                                                /* 225 - aacute */
        "a",                                                 /* 226 - acirc */
        "a",                            /* 227 - atilde (approx unaccented) */
        "a",                                                  /* 228 - auml */
        "a",                                                 /* 229 - aring */
        "ae",                                                /* 230 - aelig */
        "c",                                                /* 231 - ccedil */
        "e",                                                /* 232 - egrave */
        "e",                                                /* 233 - eacute */
        "e",                                                 /* 234 - ecirc */
        "e",                                                  /* 235 - euml */
        "i",                                                /* 236 - igrave */
        "i",                                                /* 237 - iacute */
        "i",                                                 /* 238 - icirc */
        "i",                                                  /* 239 - iuml */
        INV_CHAR,                                        /* 240 - eth (n/a) */
        "n",                                                /* 241 - ntilde */
        "o",                                                /* 242 - ograve */
        "o",                                                /* 243 - oacute */
        "o",                                                 /* 244 - ocirc */
        "o",                            /* 245 - otilde (approx unaccented) */
        "o",                                                  /* 246 - ouml */
        INV_CHAR,                                           /* 247 - divide */
        "o",                            /* 248 - oslash (approx unaccented) */
        "u",                                                /* 249 - ugrave */
        "u",                                                /* 250 - uacute */
        "u",                                                 /* 251 - ucirc */
        "u",                                                  /* 252 - uuml */
        "y",                            /* 253 - yacute (approx unaccented) */
        INV_CHAR,                                      /* 254 - thorn (n/a) */
        "y"                                                   /* 255 - yuml */
    };

    /*
     *   Map certain extended characters into our 128-255 range.  If we
     *   don't recognize the character, return the default invalid
     *   charater value.  
     */
    if (html4_char > 255)
    {
        switch(html4_char)
        {
        case 338: html4_char = 140; break;
        case 339: html4_char = 156; break;
        case 376: html4_char = 159; break;
        case 352: html4_char = 154; break;
        case 353: html4_char = 138; break;
        case 8211: html4_char = 150; break;
        case 8212: html4_char = 151; break;
        case 8216: html4_char = 145; break;
        case 8217: html4_char = 146; break;
        case 8218: html4_char = 130; break;
        case 8220: html4_char = 147; break;
        case 8221: html4_char = 148; break;
        case 8222: html4_char = 132; break;
        case 8224: html4_char = 134; break;
        case 8225: html4_char = 135; break;
        case 8240: html4_char = 137; break;
        case 8249: html4_char = 139; break;
        case 8250: html4_char = 155; break;
        case 8482: html4_char = 153; break;

        default:
            /* unmappable character */
            result[0] = (unsigned char)INV_CHAR;
            result[1] = '\0';
            return;
        }
    }
    
    /* 
     *   if the character is in the regular ASCII zone, return it
     *   untranslated 
     */
    if (html4_char < 128)
    {
        result[0] = (unsigned char)html4_char;
        result[1] = '\0';
        return;
    }

    /* look up the character in our table and return the translation */
    strcpy(result, xlat_tbl[html4_char - 128]);
}
